import java.io.IOException;
import java.util.Properties;
import BankApp.*;

import org.omg.CORBA.ORB;

import BankApp.Packet;
import BankServer.UDPClientSocket;
import BankServer.UDPServerSocket;
import ServerConfig.ServerConfig;

/**
 * @author Min Huang (khtwo)
 * BankMonitor is used to accept request of relaunch a server. Monitors will periodically check
 * each other. When a monitor is found failed, it will be launched automatically by other monitors.
 * This monitor can also check servers status in a bigger time cycle(about 7.5 seconds once).
 */
public class BankMonitor {
	public static UDPClientSocket sender = new UDPClientSocket();
	public static ServerConfig currentConfig = new ServerConfig();	// Current configure of this server
	public static UDPServerSocket UDPServer = null;	// The socket to listen request
	public static ORB orb;				// The orb
	public static boolean programEnd = false;
	public static String[] startArgs = null;

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		startArgs = args;
		// Add a hook to handle exit
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				// System.out.println("System shutting down.");
				programEnd = true;
				
				// Close the UDP server socket if it remains open
				if (UDPServer != null)
					// if (!UDPServer.isClosed())
					UDPServer.close();


				// test
				if(BankUtility.DEBUG_LEVEL > 1){
					try{
						System.out.println("Press any key to end");
						System.in.read();
					}catch(Exception e){
					}
				}
						
			}
		});

		// Check if the sender UDP socket is working
		if (sender == null){
			System.out.println("No UDP socket is usable.");
			Runtime.getRuntime().exit(1);
		}

		// Using specific working directory
		if(args.length>0){
			currentConfig.centralNSConfigFileName = args[0] + "\\" + currentConfig.centralNSConfigFileName;
			currentConfig.serverConfigFileName = args[0] + "\\" + currentConfig.serverConfigFileName;
			currentConfig.monitorConfigFileName = args[0] + "\\" + currentConfig.monitorConfigFileName;
			currentConfig.currentDirectory = args[0];
		}
		// Load configure from centralNS.cfg, monitor.cfg
		loadConfig();
		// Register the monitor on central naming service
		registerMonitor();

		// Run the monitor
		runMonitor();
	}

	/**
	 * Load configure from centralNS.cfg, monitor.cfg
	 */
	public static void loadConfig(){
		// Try to read monitor config
		try {
			currentConfig.readConfigFromFile(false);
		}
 		catch (Exception e) {
			// No config is read, do nothing and quit
			System.out.println("Reading configure failed. \nPlease check if centralNS.cfg, server.cfg or monitor.cfg exists.");
			Runtime.getRuntime().exit(1);
		}
		
		// Retrieve the central naming service
		Properties props = new Properties();
		props.put("org.omg.CORBA.ORBInitialPort", ServerConfig.centralNCInitialPort);
		props.put("org.omg.CORBA.ORBInitialHost", ServerConfig.centralNCInitialHost);
		String empty[] = new String[0];
		orb = ORB.init(empty, props);
		currentConfig.setServerConfig(orb);
	} // End of loadConfig

	/**
	 * Register the monitor on central naming service
	 */
	public static void registerMonitor(){
		// Check if this monitor has been registered on central naming service
		try{
			ServerConfig.getMonitorConfig(currentConfig.rootContext, currentConfig.monitorID);
		}catch(Exception e){
			// This monitor has not been registered, register it on central naming service
			try {
				System.out.println("The monitor information is not found on central naming service.\nRegistering the server...  ");
				Runtime.getRuntime().exec("java ServerConfig.RegisterProgramLoader").waitFor();
				System.out.println("Monitor registration OK.");
			}
			catch (Exception e1) {
				// Registration failed, quit
				System.out.println("Monitor registration failed.");
				Runtime.getRuntime().exit(1);
			}
		}
	} // End of registerServer

	public static void runMonitor(){
		try{
			BankMonitor.UDPServer = new UDPServerSocket(Integer.parseInt(currentConfig.UDPPort));
		}catch(NumberFormatException e){
			System.out.println("Bad configure found in current configure.");
			System.exit(1);
		}catch(Exception r){
			// The UDP port is taken or another instance of server is running, quit current execution
			System.out.println("Another instance of server is running, quit.");
			Runtime.getRuntime().exit(2);
		}
		
		System.out.println("Monitor on "+currentConfig.localhostIP+":"+currentConfig.UDPPort+" started.");
		
		MonitorUDPListener service = new MonitorUDPListener();
		service.start();

		// Repeat monitoring the servers
		int serverMonitorWait = 0, selfInfoMissingRetry = 0;
		while(!programEnd){
			// Scan current running servers and monitors
			try{
				// Check if the monitor registration is still in the naming service
				try{
					ServerConfig myself = ServerConfig.getMonitorConfig(currentConfig.rootContext, currentConfig.monitorID);
					selfInfoMissingRetry = 0;
				}catch(Exception e){
					if(selfInfoMissingRetry++ > 5){
						// Not found in the central naming service, shutdown itself
						programEnd = true;
						continue;
					}
				}
				
				
				String[] runningMonitors = ServerConfig.listMonitors(currentConfig.rootContext);
				// Test monitors, restart them when one of them failed
				for(int i=0; i<runningMonitors.length; i++){
					// Retrieve a config of the monitor
					ServerConfig monitorConfig = ServerConfig.getMonitorConfig(currentConfig.rootContext,
							runningMonitors[i].substring(ServerConfig.monitorTag.length(), ServerConfig.monitorTag.length()+1));
					try{
						// Do not test it self
						if(monitorConfig.monitorID.compareTo(currentConfig.monitorID)!=0){
							Packet testPacket = new Packet();
							testPacket.status = Packet.TEST;
							sender.sendPacket(monitorConfig.localhostIP, Integer.parseInt(monitorConfig.UDPPort),testPacket);
						}
					}catch(Exception e){
						// Can not contact with remote listener, remote failed, try to relaunch it
						try{
							System.out.println("Found monitor" + monitorConfig.monitorID + " failed, relaunch it.");
							monitorConfig.relaunch();
						}catch(Exception er){}
					}
				}

				// Test Bank Server after several times of monitor test
				serverMonitorWait++;
				if(serverMonitorWait>5){
					serverMonitorWait=0;
					String[] runningServers = ServerConfig.listServers(currentConfig.rootContext);
					// Test Servers, restart them when one of them failed
					for(int i=0; i<runningServers.length; i++){
						String branchID = runningServers[i].substring(6,9);
						String serverID = runningServers[i].substring(16,runningServers[i].length());
	
						// Retrieve a config of the server
						ServerConfig server = ServerConfig.getServerConfig(currentConfig.rootContext, branchID, serverID);
						try{
							Packet testPacket = new Packet();
							testPacket.status = Packet.TEST;
							sender.sendPacket(server.localhostIP, Integer.parseInt(server.backupUDPPort),testPacket);
						}catch(Exception e){
							// Can not contact with remote listener, remote failed, try to relaunch it
							try{
								System.out.println("Found server" + server.branchID + "\\" + server.serverID + " failed, relaunch it.");
								server.relaunch();
							}catch(Exception er){}
						}
					}
				}

			}catch(Exception e){
				// Error accessing naming service, retry
				
				/* Depreciated*/
				System.out.println("Accessing central naming service failed, quit.");
				programEnd = true;
				Runtime.getRuntime().exit(2);
				/**/
			}

			// Sleep for a while
			try{
				Thread.sleep(1500);
			}catch(Exception e){}
			
		}
	}
	
	/**
	 * @author Min Huang (khtwo)
	 * The monitor service to listen to relaunch request
	 */
	static class MonitorUDPListener extends Thread{
		public void run(){
			Packet receivedRequest = null;
			// Check if the UDP server socket working
			if(BankMonitor.UDPServer==null){
				System.out.println("Server socket failed");
				return;
			}
			
			// Listen to the request from other clients and process it
			while(!programEnd){
				try{
					receivedRequest = BankMonitor.UDPServer.recvPacket();
					// Ignore test packet
					if(receivedRequest.status == Packet.TEST || receivedRequest.status == Packet.TEST_REPLY) continue;
					
					if(receivedRequest.status == Packet.RELOAD){
						// The request is to relaunch a program
						String branchID = receivedRequest.opID.trim();
						System.out.println("Received request of relaunching server in branch"+branchID);
						RelaunchThread launch=new RelaunchThread();
						launch.setDaemon(true);
						launch.loadingServer = true;
						launch.branchID = branchID;
						launch.isPrimary = false;
						launch.start();
					}
					
				}catch(Exception e){
					// Time out, do nothing
				}
			}
			BankMonitor.UDPServer.close();
		}
	}
	
	/**
	 * @author Min Huang (khtwo)
	 * The thread to relaunch a server or a monitor
	 */
	static class RelaunchThread extends Thread{
		boolean loadingServer = true;
		String branchID = null;
		boolean isPrimary = false;  
		String monitorID = null;
		
		public void run(){
			if(loadingServer){
				// Load a branch server
				try{
					ServerConfig lostServer = ServerConfig.getServerConfig(currentConfig.rootContext, branchID, isPrimary);
					lostServer.relaunch();
				}catch(Exception e){}
			}
			else{
				// load a monitor
				try{
					ServerConfig lostServer = ServerConfig.getMonitorConfig(currentConfig.rootContext, monitorID);
					lostServer.relaunch();
				}catch(Exception e){}
			}
		}
	}
}
