/******************************************************************
1. I upload a new version of UDP in folder "apr2".
2. It now can accept arbitary number of connection (<=256). It take advantages of "DatagramSocket" which uses a fixed local 
	port number when sending out packet. So, hostname + portnumber becomes a unique connection ID for receiver to maintain the sequence number. 
3. I also change the stop-and-wait algorithms a little bit. Now the sender increment sequence number when sending out is successful. 
The receiver does like this:
a) if the packet is from a new connection, record the sender's sequence number and add this connection into table.
b) if the packet is from an old connection, compare the sequence number with previous one. If it is bigger than previous, update sequence number, 
	acknowledge message. If it is smaller than or equal to previous message, ignore.

4. My design is based on an assumption of usage in our "primary-primary" transfer operation such that to use a fixed "sending" socket for 
	each remote primary server. By doing so, the socket's local port number will remain unchanged.  i.e. we have primary servers A,B,C. Then server 
	A will use two client socket b,c to connect with server B,C for transfer request so that the port number of b,c will be a fixed as long as 
	connection is not closed. 


*****************************************************************/


package BankServer;

import java.net.*;
import java.io.*;
import BankApp.*;

//import java.io.ByteArrayInputStream;
//import java.io.DataInputStream;




//this socket should be used in such a way that it can either be a client socket which send
//out message or as server socket, but it cannot be both. 
public class UDPSocket 
{	

	public int replyPortNumber;
	public String replyHostName;
	
	protected long seqNumber[]=new long[BankUtility.MaxUDPConnectionNumber];
	protected String hostNames[]=new String[BankUtility.MaxUDPConnectionNumber];
	protected int portNumbers[]=new int[BankUtility.MaxUDPConnectionNumber];
	protected int connectionCount=0;

	protected int myPortNumber;
	protected DataInputStream dataIn;
	protected DataOutputStream dataOut;
	protected ByteArrayInputStream byteIn;
	protected ByteArrayOutputStream byteOut;
	protected DatagramPacket recvPacket, sendPacket;
	protected DatagramSocket recvSocket, sendSocket;
	protected byte buffer[];

	public void close() 
	{
		if (recvSocket != null)
			if (!recvSocket.isClosed())
				recvSocket.close();
		if (sendSocket != null)
			if (!sendSocket.isClosed())
				sendSocket.close();
	}

	public int getIndexByName(String theHostName, int portNumber)
	{
		for(int i=0; i<connectionCount; i++)
		{			
			if (hostNames[i].compareTo(theHostName)==0&&portNumber==portNumbers[i])
			{
				return i;
			}
		}			
		return -1;//in case we have error,	
	}

	public UDPSocket()
	{
		//BankUtility.SocketPacketHeadSize
		buffer=new byte[BankUtility.PacketSize];
		for (int i=0; i<BankUtility.PacketSize; i++)
		{
			buffer[i]=2;
		}
		byteIn=new ByteArrayInputStream(buffer);
		dataIn=new DataInputStream(byteIn);
		byteOut=new ByteArrayOutputStream(BankUtility.PacketSize);
		dataOut=new DataOutputStream(byteOut);
		recvPacket=new DatagramPacket(buffer, BankUtility.PacketSize);
		for (int i=0; i<BankUtility.MaxUDPConnectionNumber; i++)
		{
			seqNumber[i]=-1;
		}	
	}

//////////////////////////////////////////////////////////////////////////////////////////////////
	//host recover must call this method at both client and server socket
	//this is the host name index and must be called whenever the corresponding host is down
	public void resetSeqNumber(int index)
	{
		seqNumber[index]=0;
	}
/////////////////////////////////////////////////////////////////////////////////////////////////////

	//attention: here is a potential risk because I didn't initialize socket in base class constructor
	//because I want to leave this at derived class
	protected Packet doRecvPacket() throws IOException, SocketException, Exception
	{
		Packet packet=new Packet();
		//byteIn.reset();		
		//recvSocket.disconnect();
		
		
		//recvPacket.setData(buffer);
		//buffer=new byte[BankUtility.PacketSize];
		//byteIn=new ByteArrayInputStream(buffer);
		//dataIn=new DataInputStream(byteIn);
		//recvPacket=new DatagramPacket(buffer, BankUtility.PacketSize);
		//Thread.sleep(2000);
		//System.out.println("do recvPacket begins");
		
		recvSocket.receive(recvPacket);	
		//byteIn=new ByteArrayInputStream();
		//String str=new String(recvPacket.getData());
		//System.out.println("now I read data directly "+str);
		
		
		// test
		// System.out.println("\nrecvPacket recieves "+recvPacket.getLength() +" bytes of data\n");

		//String str=new String(recvPacket.getData());
		//System.out.println("preview received data in doRecvPacket"+str);
		
		byte temp[]=new byte[BankUtility.OpIDLength];	
			
		//System.out.println("before read fully to temp");

		dataIn.readFully(temp, 0,  BankUtility.OpIDLength);//BankUtility.OpIDLength
		//System.out.println("after read fully to temp");
		packet.opID=new String(temp);		
		packet.seqNumber=dataIn.readLong();
		packet.opType=dataIn.readInt();
		packet.source=dataIn.readInt();
		packet.dest=dataIn.readInt();
		packet.oldValue=dataIn.readFloat();
		packet.newValue=dataIn.readFloat();
		packet.tag=dataIn.readInt();
		packet.status=dataIn.readInt();
		
		//System.out.println("before doRecvPacket return ");
		//packet.display();
		byteIn.reset();
		//recvSocket.close();

		// debug
		BankUtility.showDebugInfo(2,"Received UDP packet from " + recvPacket.getAddress().getHostAddress() + ":" + recvPacket.getPort());
		BankUtility.showDebugInfo(3,"",packet);

		return packet;
	}	
	
	//attention: here is a potential risk because I didn't initialize socket in base class constructor
	//because I want to leave this at derived class	
	public void doSendPacket(String hostName, int portNumber, Packet packet) throws IOException
	{		
		InetAddress iNet=InetAddress.getByName(hostName);
		if(packet.opID == null){
			packet.opID = BankUtility.defaultOpID;
		}else if(packet.opID.length()<BankUtility.OpIDLength){
			packet.opID = packet.opID + BankUtility.defaultOpID.substring(0,BankUtility.OpIDLength-packet.opID.length());
		}else if(packet.opID.length()>BankUtility.OpIDLength){
			packet.opID = packet.opID.substring(0,BankUtility.OpIDLength);			
		}

		dataOut.writeBytes(packet.opID);		
		dataOut.writeLong(packet.seqNumber);
		dataOut.writeInt(packet.opType);
		dataOut.writeInt(packet.source);
		dataOut.writeInt(packet.dest);
		dataOut.writeFloat(packet.oldValue);
		dataOut.writeFloat(packet.newValue);
		dataOut.writeInt(packet.tag);
		dataOut.writeInt(packet.status);
		//datagrampacket is newed 
		sendPacket=new DatagramPacket(byteOut.toByteArray(), BankUtility.PacketSize, iNet, portNumber);//
		//String str=new String(sendPacket.getData());
		//System.out.println("before send out packet, I read data directly "+str);
		sendSocket.send(sendPacket);
		// debug
		BankUtility.showDebugInfo(2,"Send UDP packet to " + iNet.getHostAddress() + ":" + portNumber);
		BankUtility.showDebugInfo(3,"",packet);
		
		byteOut.reset();
	}
}
